from __future__ import nested_scopes
__docformat__ = 'reStructuredText'

# Python imports
import StringIO

# Zope imports
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager

from ZODB.POSException import ConflictError

# Plone imports
from Products.CMFCore.utils import getToolByName
from Products.CMFCore import CMFCorePermissions
from Products.Archetypes.public import *
from Products.Archetypes.Field import ObjectField, encode, decode, registerField

# Local product imports
from Products.DataGridField import DataGridWidget
from Products.DataGridField.Column import Column
from Products.DataGridField.interfaces import IDataGridField

class DataGridField(ObjectField):
    """ Table field with undefined number of rows
    
    DataGridField provides an user fillable table with fixed 
    columns and undefined number of rows.
    
    Data is maintained internally:
        - DataGridField.value is a list 
        - Each list item represents a row
        - Each row is a dictionary using column names as a key        
    
    DataGridField properties:
        - searchable:
            - If true all the contents of the DataGridField is concatenated
              to searchable text and given to text indexer
        - columns
            - Sequence of column names             
        - widget:
            - must be DataGridWidget
    
    """
    __implements__ = (ObjectField.__implements__, IDataGridField,)

    _properties = ObjectField._properties.copy()
    _properties.update({
        'type' : 'datagrid',
        'mode' : 'rw',
        'default' : ({},),
        'widget' : DataGridWidget,
        'columns' : ('default',),
        })

    security = ClassSecurityInfo()
    
    
    def __init__(self, name=None, **kwargs):
        """ Create DataGridField instance
        """ 
        
        # call super constructor
        ObjectField.__init__(self, name, **kwargs)
        
    def getColumnIds(self):
        """ Return list of column ids """        
        return self.columns
        
    security.declarePrivate('set')
    def set(self, instance, value, **kwargs):
        """
        The passed in object should be a records object, or a sequence of dictionaries
        
        
        
        """
        __traceback_info__ = value, type(value)
        
        # we sanitize the values
        cleaned = []
        for row in value:
            val = {}
            empty = 1
            for col in self.getColumnIds():
                val[col] = (row.get(col,'')).strip()
                if val[col]:
                    empty = 0
            if not empty:    
                cleaned.append(val.copy())
                
                    
        #value = [decode(v.strip(), instance, **kwargs)
        #         for v in value if v and v.strip()]
        value = cleaned[:]
        value = tuple(value)                
        
        # fill in data        
        ObjectField.set(self, instance, value, **kwargs)

    security.declarePrivate('get')
    def get(self, instance, **kwargs):
        """ Return DataGridField value
        
        Value is a list object of rows.
        
        If parameter mimetype == 'text/plain' is passed, 
        a string containing all cell values concatenated together is returned.
        This is for site indexing services (DataGridField.searchable = true).
        """
        
        if(kwargs.has_key('mimetype') and kwargs['mimetype'] == 'text/plain'):
            # Data is returned for text indexing
            # Concatenate all cell values 
            buffer = StringIO.StringIO()
            
            value = ObjectField.get(self, instance, **kwargs) or ()
            for row in value:
                for col in self.getColumnIds():
                    buffer.write(row.get(col,''))
                    # separate the last word of a cell 
                    # and the first of the next cell
                    buffer.write(' ')
            
            return encode(buffer.getvalue(), instance, **kwargs)
            
        else:
            # Return list containing all encoded rows            
            value = ObjectField.get(self, instance, **kwargs) or ()
            data = [encode(v, instance, **kwargs) for v in value]

            return tuple(data)

    security.declarePrivate('getRaw')
    def getRaw(self, instance, **kwargs):
        return self.get(instance, **kwargs)

    security.declarePublic('get_size')
    def get_size(self, instance):
        """Get size of the stored data used for get_size in BaseObject
        """
        size=0
        for line in self.get(instance):
            size+=len(str(line))
        return size
    
    # Grid API as defined in interfaces.py
    
    security.declarePublic('getColumns')
    def getColumns(self, instance):
        return self.columns
        
    security.declarePublic('getRowCount')
    def getRowCount(self, instance):
        return len(self.get(instance))
    
    def getRow(self, instance, rowNumber):
        """Get the given row by number. The first row is rowNumber 0. Returns 
        a dict of the row.
        """
        data = self.get(instance)
        return data[rowNumber]

    def getColumn(self, instance, columnName):
        """Get data in the given column by name as a tuple of values.
        """
        data = self.get(instance)
        col = []
        for row in data:
            col.append(row[columnName])
        return tuple(col)
        
    def search(self, instance, key=None, **kwargs):
        """Search for rows. If key is given look for this value in the first 
        column. If key is not given, at least one kwarg must be given, 
        specifying column names and values. For example, 
            search(surname='Jones', title='Mr') 
        will return all rows with the surname and title matching Mr. Jones.
        Always returns a tuple of dicts.
        """
        data = self.get(instance)
        if key is not None:
            kwargs[self.getColumnIds()[0]] = key
        matches = []
        for r in data:
            match = True
            for k, v in kwargs.items():
                if r[k] != v:
                    match = False
            if match:
                matches.append(r)
        return tuple(matches)
                    
        
    def lookup(self, instance, key, column, lookupColumn=None):
        """Look for the given key in the column specified by lookupColumn.
        If no lookupColumn is given, look in the first column. For the first row
        found, return the value stored in the corresponding column as given by
        the 'column' parameter. Returns None if the key could not be found. 
        
        (if you've used Excel, this is similar to VLOOKUP)
        """
        data = self.get(instance)
        if lookupColumn is None:
            lookupColumn = self.columns[0]
        for r in data:
            if r[lookupColumn] == key:
                return r[column]
        return None
        

    def getAsDisplayList(self, instance, keyCol=None, valueCol=None):
        """Get two columns of each row as a DisplayList - the key columns is
        keyCol, and the value column is valueCol. If these are not given,
        use the first two columns, respectively.
        """
        data = self.get(instance)
        
        if keyCol is None:
            keyCol = self.getColumnIds()[0]
        if valueCol is None:
            valueCol = self.getColumnIds()[1]
        
        lst = DisplayList()    
        for r in data:
            lst.add(r[keyCol], r[valueCol])
            
        return lst
        
    def getAsGrid(self, instance):
        """Return a tuple of tuples - the outer tuple has one element
        per row in the grid, the inner tuple has one element per column
        in that row.
        """
        data = self.get(instance)
        rows = []
        for r in data:
            rows.append(tuple([r[c] for c in self.getColumnIds()]))
        return tuple(rows)

registerField(DataGridField,
              title='DataGridField',
              description=('Used for storing tabular string data'))
